/**
 * Example of how to call the TCB DistributedFailureDetection service.
 * 
 * This client demo is responsible for starting a distributed failure detection
 * (and the associated measurement) that is ended by server demo. 
 * Only server demo presents the associated measurement results.
 * 
 * A thread is launched to monitor the results of distributed timing 
 * failure detection.
 * 
 * To call this service, you need first to create a DLL with
 * two functions:
 * 
 * 		1) The function to timely execute if a failure occurs
 * 		2) A function that returns the Worst Case Execution Time
 * 		   (milliseconds) of the function specified in 1). 
 * 
 * There is already a dll example created that works with this example.
 * The source code (Embedded Visual C++ 3.0) is located in directory
 * "tcb_services_examples/dll_example_for_tcb_handler". You should compile it
 * and deploy in PocketPC prior to running this example.
 * 
 * @author Paulo Sousa
 */

#include "stdafx.h"
//include sockets header file
#include <winsock.h>
//include tcb services library header file
#include "..\..\..\tcb_services_library\tcb_services_library.h"


//TODO: replace this values according to your situation
#define MY_IP "192.168.0.1"
#define SERVER_IP "192.168.0.2"
#define SERVER_PORT 40000


/**
 * Task that prints 1 distributed FD info sent by TCB and exits
 */
DWORD WINAPI wait_info_task (LPVOID param) {
	tcb_distributed_fd_info_t info;
	
	//wait for distributed FD info
	info = tcb_wait_info();
	if (info.from==-1) {
		printf("Error: TCB session is not opened\n");
		return 0;
	}
	//print info sender and sequence number
	printf("W_INFO:Info from %d, seq=%d\n", info.from, info.seq);
	//print fail info for each machine (per tcb id)
	for (int i=0; i<NUMBER_OF_MACHINES; i++) {
		//if there exists no info
		if (info.failInfo[i].fail==2) {
			printf("W_INFO:Id %d: NO INFO\n", i);
		}
		//if there exists info
		else {
			printf("W_INFO:Id %d: ",i);
			//verify if a timing failure occured
			if (info.failInfo[i].fail)
				printf("failed");
			else printf("timely");
			//if lateness is available
			if (info.failInfo[i].latenessAvail)
				printf(", lat.degree=%d ms\n", info.failInfo[i].latenessDegree);
			else printf(", no lateness available\n");
		}
	}
	
	return 0;
}




int main(int argc, char* argv[]) {

	//open_session
	printf("Opening TCB session\n");
	int error = open_tcb_session();
	if (error!=0) {
		printf("Error opening session=%d\n",error);
		getchar();
		return -1;
	}

	//launch wait_info thread
	HANDLE task_id =
		CreateThread(NULL,0,wait_info_task,NULL,NULL,NULL);
	if (task_id==NULL){
		printf("FATAL ERROR: Couldn't start wait_info task\n");
		getchar();
		return -1;
	}

	//create a TCB socket
	printf("Creating TCB socket\n");
	error = tcb_socket(AF_INET);
	if (error!=0) {
		printf("Error creating TCB socket: %d\n", error);
		getchar();
		return -1;
	}

	//bind TCB socket
	printf("Binding TCB socket\n");
	struct sockaddr_in sa;
	sa.sin_family=AF_INET;
	sa.sin_addr.s_addr= inet_addr(MY_IP);
	sa.sin_port=htons(0);
	error=tcb_bind((struct sockaddr*)&sa, sizeof(sa));
	if (error!=0) {
		printf("Error binding TCB socket: %d\n", error);
		getchar();
		return -1;
	}

	//start a distributed FD
	printf("Starting a distributed FD\n");
	//always reserve TCB_DATA_SIZE bytes in the end of the buffer
	char buf[50+TCB_DATA_SIZE];
	//msg to be sent
	strcpy(buf, "Hi server!(FD)");
	struct sockaddr_in to;
	to.sin_family = AF_INET;
	to.sin_addr.s_addr = inet_addr(SERVER_IP);
	to.sin_port = htons(SERVER_PORT);
	//recipients: TCB with id 1 (0x2 = byte with bit 1 set)
	char dest_mask = 0x2;
	//time to failure (ms)
	int spec = 15;
	//maximum time allowed for the execution of failure handler (ms)
	int deadline = 0;
	char dllName[100];
	char funcName[100];
	char wcetName[100];
	//DLL where resides the failure handler and the wcet function
	strcpy(dllName, "tcb_handler_example.dll");
	//failure handler name
	strcpy(funcName, "distFD_example");
	//wcet name
	strcpy(wcetName, "distFD_example_wcet");
	
	int nbytes = tcb_sendtoWithFD(buf, 50+TCB_DATA_SIZE, NULL, (struct sockaddr*)&to, 
						    sizeof(to), get_tcb_timestamp(),
							dest_mask, spec, deadline, dllName, funcName, wcetName);
		
	if (nbytes<0) {
		printf("Error starting distributed FD: %d\n", nbytes);
	}
	else printf("%d bytes sent with FD\n", nbytes);


	//wait for wait_info thread
	WaitForSingleObject(task_id, INFINITE);

	//close_session
	printf("Closing session\n");
	error=close_tcb_session();
	if (error!=0) {
		printf("Error closing session=%d\n",error);
		getchar();
		return -1;
	}
	printf("Close ok!\n");

	//wait for user to press a key
	printf("Press a key to close window...\n");
	getchar();
	return 0;
}
