/**
 * Header file of TCB services C library.
 * This header file defines the structs and functions not exported.
 *
 * author: Paulo Sousa (pjsousa@di.fc.ul.pt)
 * date: 21/03/2003
 */

//stopMeasurement return value
typedef struct {
	int end_ev;
	int procdelay;
} measurement_info_t;


//used in fail_info_to_app. Stores info about a failure.
typedef struct fail_info {
	//0-no failure, 1-failure, 2-no info
	char fail;
	//lateness is not available (FALSE), or available(TRUE)
	bool lateness_avail;
	//lateness degree if avaialable
	int lateness_degree;
} fail_info_t;

//for memory mapping between TCB DistFD service and 
//wait_info() function implemented in TCB DLL
typedef struct fail_info_to_app {
	char broadcast;
	//msg sender id (tcb id)
	char from;
	//msg sequence number (FD context)
	int seq;
	//info about failure in each machine
	fail_info failure_info[NUMBER_OF_MACHINES];
} fail_info_to_app_t;		


//for memory mapping between TCB services DLL and TCB
#define GET_TIMESTAMP 0
#define START_MEASUREMENT 1
#define END_MEASUREMENT 2
#define TIMELY_EXEC 3
#define START_LOCAL_FD 4
#define END_LOCAL_FD 5
#define GET_TCB_ID 6
#define START_DISTRIBUTED_FD 7
#define MSG_RCV_ON_PAYLOAD 8
#define GET_GLOBAL_TIMESTAMP 9
#define DEFINE_REMOTE_HANDLER 10
#define START_DISTRIBUTED_REMOTE_FD 11

typedef struct service_request_response {
	//0-empty, 1-service request, 2- service response
	char type;
	union {
		//request
		struct {
			//defines above
			char service_id;
			//service input parameters
			union {
				//getGlobalTimestamp input value
				int localInstant;
				//startMeasurement input value
				int start_ev;
				//endMeasurement and endLocalFD input value
				int tag;
				//timelyExec input value
				struct {
					//a time instant
					int start_ev; 
					//start_ev+delay marks the exec begin
					int delay; 
					//deadline for the execution, counting since start_ev
					int max_exec; 
					//DLL file where the function is defined
					wchar_t dllName[50]; 
					//function name
					wchar_t funcName[50]; 
					//Worst Case Execution Time function name
 					wchar_t wcetName[50]; 
				} execArguments;
				//startLocalFD input value
				struct {
					//start instant of the observed execution
					int start_ev;
					//specified execution duration
					int spec;
					//maximum time allowed for the execution of the handler
					//(if a failure occurs)
					int deadline;
					//DLL file where the handler is defined
					wchar_t dllName[50]; 
					//handler name
					wchar_t funcName[50]; 
					//handler's Worst Case Execution Time function name
 					wchar_t wcetName[50]; 
				} localFDArguments;
				//startDistributedFD and (partially)startDistributedRemoteFD input value
				struct {
					//bit mask indicating packet destinations
					wchar_t destMask;
					//start instant of the observed execution
					int send_ev;
					//specified execution duration
					int spec;
					//maximum time allowed for the execution of the handler
					//(if a failure occurs)
					int deadline;
					//DLL file where the handler is defined
					wchar_t dllName[50]; 
					//handler name
					wchar_t funcName[50]; 
					//handler's Worst Case Execution Time function name
 					wchar_t wcetName[50]; 
				} distFDArguments;
				//msgRcvOnPayload input value
				struct {
					int seq;
					wchar_t from;
					int time_of_receipt;
				} msgRcvOnPayloadArguments;
				//defineRemoteHandler input value
				struct {
					//DLL file where the handler is defined
					wchar_t dllName[50]; 
					//handler name
					wchar_t funcName[50]; 
					//handler's Worst Case Execution Time function name
 					wchar_t wcetName[50]; 
					//maximum time allowed for the execution of the handler
					//(if a failure occurs)
					int deadline;
				} remote_handler;
			};
		} request;

		//response
		struct {
			//defines above
			char service_id;
			//service return values
			union {
				//getTimestamp and getGlobalTimestamp return value	
				int value;
				//startMeasurement and startLocalFD return value	
				int tag;
				//endMeasurement return value
				measurement_info_t mi;
				//timelyExec return value
				int end_ev;
				//endLocalFD return value
				struct {
					//when the detection was turned off
					int end_ev;
					//end_ev - start_ev
					int procdelay;
					//a timing failure ocurred?
					bool faulty;
				} endLocalFDRet;
				//getTCBId return value
				int id;
				//startDistributedFD return value
				int seq;
				//msgRcvOnPayload return value
				bool deliver;
				//defineRemoteFDHandler return value 
				int ret;
			};
		} response;
	};
} service_request_response_t;

//stores info associated with the TCB socket (only exists one simultaneous TCB socket)
typedef struct tcb_socket_info {
	//real socket descriptor
	int socket_fd;
	//tcb id
	char tcb_id;
} tcb_socket_info_t;
tcb_socket_info_t tcb_s_info;


//offset of each type of information that goes in each packet sent through a TCB socket
int SEND_EV_OFFSET = 0;
int SEQ_OFFSET = 4;
int TCB_ID_OFFSET = 8;

/**
 * Distributed measurement error (ms)
 */
#define DISTRIBUTED_MEASUREMENT_ERROR 6


/**
 * Calls the tcb service: start_distributed_fd
 * 
 * Starts a distributed timing failure detection.
 * 
 * @param destMask Bit mask indicating packet destinations.
 * 					if (bit n == 1) then pack is sent to machine n.
 * 					E.g. destMask=0x7 => send to machines 0,1,2
 * @param sendEv The initial time instant of the failure detection.
 * @param spec Maximum duration starting at sendEv.
 * @param deadline Maximum exec time of failure handler (if a failure occurs)
 * @param dllName Name of the DLL file where the handler is defined
 * @param funcName Handler name 
 * @param wcetName Handler's Worst Case Execution Time function name
 *
 * @return seq that identifies the dist. timing failure detection,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection. 
 */ 
int start_tcb_distributed_fd(char destMask, int sendEv, int spec, 
						 int deadline, char* dllName, char* funcName, 
						 char* wcetName);

/**
 * Calls the tcb service: start_distributed_remote_fd
 * 
 * Starts a distributed remote timing failure detection.
 * 
 * @param destMask Bit mask indicating packet destinations.
 * 					if (bit n == 1) then pack is sent to machine n.
 * 					E.g. destMask=0x7 => send to machines 0,1,2
 * @param sendEv The initial time instant of the failure detection.
 * @param spec Maximum duration starting at sendEv.
 *
 * @return seq that identifies the dist. timing failure detection,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection. 
 */ 
int start_tcb_distributed_remote_fd(char destMask, int sendEv, int spec);



/**
 * PSEUDO-SERVICE: msg_rcv_on_payload
 *
 * GOAL: to DLL notify TCB that a msg was received on payload
 * 
 * PARAM: 
 *		seq - sequence number
 *		from - sender id
 *		time_of_receipt - reception time instant
 *
 * RETURNS : TRUE if msg can be delivered,
 *					  FALSE if msg cannot be delivered			   
 */
bool tcb_msg_rcv_on_payload(int seq, char from, int time_of_receipt);


/**
 * Sends data through TCB socket
 *
 * @param buf buffer that contains the data to be transmitted. 
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len length of buf
 * @param flags specifies the way in which the call is made
 * @param to address of the target socket
 * @param tolen length of address
 * @param int send_ev the initial time instant of the action to be measured
 *
 * @return total number of bytes sent if OK, 
 *		   SOCKET_ERROR if a socket error occurs
 */
int my_send (char* buf, int len, int flags, struct sockaddr *to, int tolen, 
			 int send_ev, int seq);

/**
	 * Desserializes a byte array containing an integer.
	 * 
	 * @param b the byte array to deserialize
	 * @param off start offset
	 * @return the deserialized integer.
	 */
	int byteArrayToInt(char* b, int off);

	/**
	 * Serializes an integer into a byte array
	 * 
	 * @param i the int to serialize
	 * @return the byte [] with the serialization
	 */
	void intToByteArray(int* i, char* b, int off);